#ifndef BackgroundIR_h
#define BackgroundIR_h

#include "Arduino.h"

// The BackgroundIR routines provide a way to send and receive raw IR signals
// at interrupt level, communicating with the main loop via a circular
// buffer.
//
// Resources used by this code:
//
//                 Arduino Uno R3      Teensy 2.0
//
// Interrupts      TIMER1_COMPA_vect   same
//                 TIMER1_CAPT_vect
//                 TIMER1_COMPB_vect
//
// Receive Timer   Timer 1             same
// Receive Pin     8 (ICP1)            22 (ICP1)
//
// Send Timer      Timer 2             Timer 4
// Send Pin        3 (OC2B)            10 (OC4A)

#include "porting.h"

// See RawIR.h for the format of raw IR codes as stashed in 16 bit words.

// All this stuff resides in the BackgroundIR namespace.

namespace BackgroundIR {

   enum IRState {
      BGIR_STOPPED,     // Receiver stopped
      BGIR_WAITING,     // Receiver started, but no data available now
      BGIR_HAS_DATA,    // One or more mark/space/status sequences available
      BGIR_SENDING,     // Queued up data is still being sent
   };

      // Call InitializeIR() during setup to setup all the timers
      // and pins used by this library to send and recieve IR. This
      // will initialize things with the receiver turned off.
      //
   void
   InitializeIR();

      // Return the current state of IR routines. While it reports
      // BGIR_HAS_DATA you can call GetData(). While it reports
      // BGIR_SENDING, it is still in the process of sending the
      // data you handed to PutData().
      //
   IRState
   GetState();

      // Enable the interrupt routine and begin listening for a
      // new IR input. (No receiving will happen unless you call this).
      // After calling this you may see GetState() return BGIR_WAITING
      // or BGIR_HAS_DATA.
      //
   void
   StartReceivingIR();

      // Disable the interrupt routine and stop listening for IR.
      // After calling this you will see BGIR_STOPPED when calling
      // GetState().
      //
   void
   StopReceivingIR();

      // Return true if the receiver is normally running by default
      // (even if it is turned off at the moment due to a send in progress).
      //
   bool
   ReceiverNormallyRunning();

      // Return IR data in format described above (undefined results
      // if you call this when GetState() != BGIR_HAS_DATA).
      //
   unsigned int
   GetData();

      // Setup the timer used for PWM output to send IR with MARK pulses
      // modulated at the specified kHz rate (rate stays the same till
      // changed, initialized to 38kHz). Input units are kHz so pass 38 to
      // mean 38000, etc.
      //
      // Return true if all OK, false if the requested rate cannot be
      // generated by the PWM output from timer2.
      //
   bool
   SetIRModulation(int kHz);

      // Return the last frequency successfully set by SetIRModulation().
      //
   int
   GetIRModulation();

      // Just check to see if PutData() would succeed, but don't push
      // anything.
      //
   bool
   PutDataOK();

      // Queue up a mark or space to be sent. Returns true if it was able
      // to push the data on the queue, false if the queue is currently
      // full and you should try again later.
      //
   bool
   PutData(unsigned int ms_data);

      // Report the size of the circular buffer compiled into the code.
      //
   int
   GetBufferSize();

      // Some "global" errors can occur if sending or receiving is
      // interrupted in the middle of the process. This routine checks
      // all the global error flags and resets them if they are set,
      // returning a bitmask of the global errors that were set.
      //
   unsigned int
   GetGlobalErrors();
};

#endif // BackgroundIR_h
